/*
 Copyright (C) 2013 Red Hat, Inc.

 This file is part of IcedTea.

 IcedTea is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published by
 the Free Software Foundation, version 2.

 IcedTea is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with IcedTea; see the file COPYING.  If not, write to
 the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 02110-1301 USA.

 Linking this library statically or dynamically with other modules is
 making a combined work based on this library.  Thus, the terms and
 conditions of the GNU General Public License cover the whole
 combination.

 As a special exception, the copyright holders of this library give you
 permission to link this library with independent modules to produce an
 executable, regardless of the license terms of these independent
 modules, and to copy and distribute the resulting executable under
 terms of your choice, provided that you also meet, for each linked
 independent module, the terms and conditions of the license of that
 module.  An independent module is a module which is not derived from
 or based on this library.  If you modify this library, you may extend
 this exception to your version of the library, but you are not
 obligated to do so.  If you do not wish to do so, delete this
 exception statement from your version.
 */
package net.sourceforge.jnlp.util.lockingfile;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.nio.file.Files;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.Map;
import java.util.WeakHashMap;
import java.util.concurrent.locks.ReentrantLock;
import net.sourceforge.jnlp.runtime.JNLPRuntime;
import net.sourceforge.jnlp.util.logging.OutputController;

/*
 * Process & thread locked access to a file. Creates file if it does not already exist.
 */
public class LockedFile {

    // The file for access
    private File file;
    private ReentrantLock threadLock = new ReentrantLock();
    private boolean readOnly;

    private final InterProcessLock processLock;

    private LockedFile(File file) {
        this.file = file;
        try {
            //just try to create
            this.file.createNewFile();
        } catch (Exception ex) {
            //intentionaly silent
        }
        if (!this.file.isFile() && file.getParentFile() != null && !file.getParentFile().canWrite()) {
            readOnly = true;
        } else {
            this.readOnly = !file.canWrite();
            if (!readOnly && file.getParentFile() != null && !file.getParentFile().canWrite()) {
                readOnly = true;
            }
        }

        if (JNLPRuntime.isWindows()) {
            processLock = new LockFile();
        } else {
            processLock = new NioFileLock();
        }
    }

    public boolean isReadOnly() {
        return readOnly;
    }
    // Provide shared access to LockedFile's via weak map
    static private final Map<File, LockedFile> instanceCache = new WeakHashMap<File, LockedFile>();

    /**
     * Get a LockedFile for a given File. Ensures that we share the same
     * instance for all threads
     *
     * @param file the file to lock
     * @return a LockedFile instance
     */
    synchronized public static LockedFile getInstance(File file) {
        if (!instanceCache.containsKey(file)) {
            instanceCache.put(file, new LockedFile(file));
        }

        return instanceCache.get(file);
    }

    /**
     * Get the file being locked.
     *
     * @return the file
     */
    public File getFile() {
        return file;
    }

    /**
     * Lock access to the file.Lock is reentrant.
     * @throws java.io.IOException
     */
    public void lock() throws IOException {
        final boolean alreadyLocked = threadLock.isHeldByCurrentThread();
        threadLock.lock();
        if (alreadyLocked) {
            // early exit: process locking may not be reentrant causing an error when trying to lock the same file twice
            return;
        }

        // Create if does not already exist, cannot lock non-existing file
        if (!isReadOnly()) {
            this.file.createNewFile();
        }

        lockProcess();
    }

    public boolean tryLock() throws IOException {
        if (this.threadLock.tryLock()) {
            processLock.lock();
            return true;
        } else {
            return false;
        }
    }

    private void lockProcess() throws IOException {
        processLock.lock();
    }

    /**
     * Unlock access to the file.Lock is reentrant. Does not do anything if not holding the lock.
     * @throws java.io.IOException
     */
    public void unlock() throws IOException {
        if (!this.threadLock.isHeldByCurrentThread()) {
            return;
        }

        try {
            if (this.threadLock.getHoldCount() == 1) {
                processLock.unlock();
            }
        } finally {
            this.threadLock.unlock();
        }
    }

    public boolean isHeldByCurrentThread() {
        return this.threadLock.isHeldByCurrentThread();
    }

    private interface InterProcessLock {
        void lock() throws IOException;

        void unlock() throws IOException;
    }

    private class NioFileLock implements InterProcessLock {
        private RandomAccessFile randomAccessFile;
        private FileChannel fileChannel;
        private FileLock fileLock;


        @Override
        public void lock() throws IOException {
            if (fileLock != null) {
                return;
            }

            if (file.exists()) {
                randomAccessFile = new RandomAccessFile(file, readOnly ? "r" : "rws");
                fileChannel = randomAccessFile.getChannel();
                if (!readOnly) {
                    fileLock = fileChannel.lock();
                }
            }
        }

        @Override
        public void unlock() throws IOException {
            if (fileLock != null) {
                fileLock.release();
            }
            fileLock = null;
            //necessary for read only file
            if (randomAccessFile != null) {
                randomAccessFile.close();
            }
            //necessary for not existing parent directory
            if (fileChannel != null) {
                fileChannel.close();
            }
        }
    }

    private class LockFile implements InterProcessLock {

        public static final double STALENESS_INTERVAL_IN_MILLIS = 2000;
        private final File lockFile = new File(file.getAbsoluteFile().getParent(), file.getName() + ".lock");

        @Override
        public void lock() throws IOException {
            if (!file.exists()) {
                return;
            }

            createParentDirIfMissing();
            if (lockFile.exists()) {
                deleteStaleLockFile();
            }
            while (!lockFile.createNewFile()) {
                try {
                    OutputController.getLogger().log(OutputController.Level.WARNING_DEBUG, "Trying to create lock file {} " + lockFile.getPath());
                    Thread.sleep(500);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
            lockFile.deleteOnExit();
        }

        private void deleteStaleLockFile() throws IOException {
            final BasicFileAttributes attr = Files.readAttributes(lockFile.toPath(), BasicFileAttributes.class);
            final long currentTime = System.currentTimeMillis();
            final long lastModifiedTime = attr.lastModifiedTime().toMillis();
            final long sinceLastModifiedInMillis = (currentTime - lastModifiedTime);
            if (sinceLastModifiedInMillis > STALENESS_INTERVAL_IN_MILLIS) {
                OutputController.getLogger().log(OutputController.Level.WARNING_DEBUG,
                        "Deleting stale lock file " + lockFile.getPath()+ " (sinceLastModifiedInMillis="+sinceLastModifiedInMillis+")");
                unlock();
            }
        }

        private void createParentDirIfMissing() {
            final File dir = lockFile.getParentFile();
            if (!dir.isDirectory()) {
                if (!dir.mkdirs()) {
                    OutputController.getLogger().log(OutputController.Level.ERROR_ALL, "failed to create parent directory of " + lockFile);
                }
                lockFile.deleteOnExit();
            }
        }

        @Override
        public void unlock() throws IOException {
            if (!lockFile.delete()) {
                if (lockFile.exists()) {
                    OutputController.getLogger().log(OutputController.Level.ERROR_ALL, "Failed to delete lock file " + lockFile);
                }
            }
        }
    }
}
